import os
import json
from contextlib import ExitStack
from pathlib import Path
from collections.abc import Iterator
from subprocess import run, call
from inspect import cleandoc
from sys import platform

from .config import Config


HOME = Path("~").expanduser()
if (HOME / ".config").exists():
    APP_PATH = HOME / ".config" / "zettel_merken"
else:
    APP_PATH = HOME / "zettel_merken"

if not APP_PATH.exists():
    os.mkdir(APP_PATH)


def get_notes_list(config: Config) -> Iterator[Path]:
    """Get a list of notes from a list of directories"""

    for dir in config.NOTE_DIRS:
        for path, dirs, files in os.walk(dir, topdown=True):

            # Exclude hidden and user ignored directories
            dirs[:] = (
                d for d in dirs if d not in config.IGNORE_DIRS and not d.startswith(".")
            )

            for file in files:
                if (
                    Path(path, file).suffix in config.INCLUDE_EXT
                    and file.casefold() not in config.IGNORE_FILES
                ):
                    yield Path(path + os.sep + file).resolve()


def get_app_path() -> Path:
    """Returns ~/.config/zettel_merken or ~/zettel_merken if .config not found"""
    return APP_PATH


def get_mail_content(notes: list[Path]) -> str:
    with ExitStack() as stack:
        files = [stack.enter_context(open(note)) for note in notes]
        return "\n\n".join(file.read() + "\n\n" for file in files)


def _add_systemd_units():
    CONF_DIR = Path("~/.config/systemd/user").expanduser().resolve()
    TIMER_UNIT = CONF_DIR / "zettel_merken.timer"
    SERVICE_UNIT = CONF_DIR / "zettel_merken.service"

    if not CONF_DIR.exists():
        os.mkdir(CONF_DIR)

    with open(TIMER_UNIT, "w") as timer:
        timer.write(
            """\
            [Unit]
            Description=Zettel Merken Daily Review Timer
            After=network-online.target

            [Timer]
            Persistent=true
            OnCalendar=Daily
            OnBootSec=300

            [Install]
            WantedBy=timers.target\
            """
        )

    with open(SERVICE_UNIT, "w") as service:
        service.write(
            """\
            [Unit]
            Description=Zettel Merken Daily Review Service

            [Service]
            Type=simple
            ExecStart=/usr/bin/python -m zettelmerken

            [Install]
            WantedBy=default.target\
            """
        )

    run("systemctl --user daemon-reload", shell=True)
    run("systemctl --user enable --now zettel_merken.timer", shell=True)


def _add_launchd_units():
    CONF_DIR = Path("~/Library/LaunchAgents").expanduser().resolve()
    TIMER_UNIT = CONF_DIR / "com.zettelmerken.dailyreview.plist"

    if not CONF_DIR.exists():
        os.mkdir(CONF_DIR)

    with open(TIMER_UNIT, "w") as timer:
        timer.write(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
            <plist version="1.0">
              <dict>
                <key>Label</key>
                <string>com.zettelmerken.dailyreview</string>
                <key>ServiceDescription</key>
                <string>Zettelmerken Daily Review</string>
                <key>ProgramArguments</key>
                <array>
                  <string>/opt/homebrew/bin/python3</string>
                  <string>-m</string>
                  <string>zettelmerken</string>
                </array>
                <key>RunAtLoad</key>
                <true />
                <key>StartCalendarInterval</key>
                <dict>
                  <key>Hour</key>
                  <integer>0</integer>
                  <key>Minute</key>
                  <integer>10</integer>
                </dict>
                <!-- For Debugging
                <key>StandardErrorPath</key>
                <string>/tmp/com.zettelmerken.dailyreview.err</string>
                <key>StandardOutPath</key>
                <string>/tmp/com.zettelmerken.dailyreview.out</string>
                -->
              </dict>
            </plist>
            """
        )

    run(f"plutil {TIMER_UNIT}", shell=True)
    run(
        f"launchctl load {TIMER_UNIT}",
        shell=True,
    )
    run(f"launchctl enable {TIMER_UNIT}", shell=True)


def add_timer_units():
    if platform == "linux":
        _add_systemd_units()
    elif platform == "darwin":
        _add_launchd_units()
    else:
        print(
            "Unrecognized platform. Adding timer units only works on linux and macos!"
        )
        exit()


def create_config():
    if not (APP_PATH / "config.json").exists():
        with open(APP_PATH / "config.json", "wt") as config:
            dump = json.dumps(
                {
                    "NOTE_DIRS": ["/path/to/notes_folder"],
                    "EMAIL": {
                        "USER": "sender@gmail.com",
                        "PASS": "sampleapppassword",
                        "HOST": "smtp.gmail.com",
                        "PORT": 465,
                    },
                    "RECEIVERS": ["receiver@gmail.com"],
                    "MAX_NOTES_PER_MAIL": 10,
                    "IGNORE_FILES": ["ignore-file.md"],
                    "IGNORE_DIRS": ["ignore-folder"],
                    "INCLUDE_EXT": [".md", ".txt", ".org", ".norg"],
                    "SCHEDULE_DAYS": [1, 3, 7, 14, 30, 60, 120],
                },
            )

            config.write(dump)


def open_config():
    default_editor = "/usr/bin/vi"  # backup, if not defined in environment vars
    path = APP_PATH / "config.json"
    editor = os.environ.get("EDITOR", default_editor)
    call([editor, path])


def show_help():
    help_str = cleandoc(
        """
        USAGE: python -m zettelmerken [OPTION]

        OPTIONS:
            --help    Show this help
            --config  Create and open config.json
            --init    Init timer units
            --remove  Remove database and systemd units

        DOC: https://www.github.com/empat94/zettel-merken.git

        NOTES:
            1. Zettel Merken uses systemd service/timer units on linux and launchd agent on macos. Timers aren't setup for windows.
        """
    )
    print(help_str)


def remove_database():
    if APP_PATH.exists():
        (APP_PATH / "zettel_merken.db").unlink(missing_ok=True)


def remove_systemd_units():
    CONF_DIR = Path("~/.config/systemd/user").expanduser().resolve()
    run("systemctl --user disable --now zettel_merken.timer", shell=True)
    run("systemctl --user disable --now zettel_merken.service", shell=True)
    (CONF_DIR / "zettel_merken.timer").unlink(missing_ok=True)
    (CONF_DIR / "zettel_merken.service").unlink(missing_ok=True)


def remove_launchd_units():
    CONF_DIR = Path("~/Library/LaunchAgents").expanduser().resolve()
    TIMER_UNIT = CONF_DIR / "com.zettelmerken.dailyreview.plist"
    run("launchctl disable com.zettelmerken.dailyreview", shell=True)
    run("launchctl unload com.zettelmerken.dailyreview", shell=True)
    (TIMER_UNIT).unlink(missing_ok=True)


def remove_timer_units():
    if platform == "linux":
        remove_systemd_units()
    elif platform == "darwin":
        remove_launchd_units()
    else:
        print("Invalid platform detected, not doing anything!")
        exit()



import unittest
import sqlite3
import json
from datetime import date
from pathlib import Path

from zettelmerken.db import Db, ScheduleExhausted, ScheduleNotFound
from zettelmerken.config import Config, ConfigNotFound
import zettelmerken.helpers as hlp


config_file = Path("extras", "config.test.json").resolve()

note1 = Path("extras", "test_folder", "note_folder_1", "note-1.md").resolve()
note2 = Path("extras", "test_folder", "note_folder_1", "note-2.txt").resolve()
note3 = Path("extras", "test_folder", "note_folder_2", "note-3.org").resolve()
note4 = Path("extras", "test_folder", "note_folder_2", "note-4.norg").resolve()
notes = [note1, note2, note3, note4]

db_path = Path("/tmp", "zettel_merken.test.db")


class TestConfig(unittest.TestCase):
    def test_get_config(self):
        config = Config.load_from_file(config_file)
        self.assertEqual(config.MAX_NOTES_PER_MAIL, 10)
        self.assertRaises(
            ConfigNotFound,
            Config.load_from_file,
            Path("extras/not_present.json"),
        )


class TestHelpers(unittest.TestCase):
    def test_get_app_path(self):
        home_dir = Path.home()
        config_dir = home_dir / ".config"
        app_dirs = [config_dir / "zettel_merken", home_dir / "zettel_merken"]
        self.assertIn(hlp.get_app_path(), app_dirs)

    def test_get_notes_list(self):
        config = Config.load_from_file(config_file)
        for note in notes:
            self.assertIn(note, list(hlp.get_notes_list(config)))


class TestDb(unittest.TestCase):
    def test_create_schema(self):
        db = Db(db_path, Config.load_from_file(config_file))
        db.create_schema()
        with sqlite3.connect(db.DB_PATH) as cx:
            cx.cursor().execute("SELECT * FROM note_schedule")

    def test_create_note_schedule(self):

        db = Db(db_path, Config.load_from_file(config_file))
        db.create_schema()
        db.create_note_schedule(note1, [1, 3, 6])
        db.create_note_schedule(note2, [0])
        db.create_note_schedule(note3, [0, 1, 2])

        with sqlite3.connect(db.DB_PATH) as cx:
            cu = cx.cursor()

            schedule1 = cu.execute(
                f"SELECT schedule FROM note_schedule WHERE note = '{note1}'"
            ).fetchone()

            assert len(json.loads(schedule1[0])) == 3

            schedule2 = cu.execute(
                f"SELECT schedule FROM note_schedule WHERE note = '{note2}'"
            ).fetchone()

            assert date.toordinal(date.today()) in json.loads(schedule2[0])

            schedule3 = cu.execute(
                f"SELECT schedule FROM note_schedule WHERE note = '{note3}'"
            ).fetchone()

            assert date.toordinal(date.today()) in json.loads(schedule3[0])

    def test_is_note_scheduled(self):
        db = Db(db_path, Config.load_from_file(config_file))
        self.assertFalse(db.is_note_scheduled(note1))
        self.assertTrue(db.is_note_scheduled(note2))
        self.assertTrue(db.is_note_scheduled(note3))
        self.assertRaises(ScheduleNotFound, db.is_note_scheduled, note4)

    def test_update_schedule(self):
        db = Db(db_path, Config.load_from_file(config_file))

        db.update_note_schedule([note2])
        self.assertRaises(ScheduleExhausted, db.is_note_scheduled, note2)

        db.update_note_schedule([note3])

        with sqlite3.connect(db.DB_PATH) as cx:
            cu = cx.cursor()

            schedule3 = cu.execute(
                f"SELECT schedule FROM note_schedule WHERE note = '{note3}'"
            ).fetchone()

            self.assertNotIn(date.toordinal(date.today()), json.loads(schedule3[0]))

    def tearDownClass():
        db = Db(db_path, Config.load_from_file(config_file))

        with sqlite3.connect(db.DB_PATH) as cx:
            cu = cx.cursor()
            cu.execute("DROP TABLE IF EXISTS note_schedule")
            cu.execute("DROP INDEX IF EXISTS idx_note")


if __name__ == "__main__":
    unittest.main()



import sys

from zettelmerken.core import main
from zettelmerken.helpers import (
    add_timer_units,
    create_config,
    open_config,
    remove_timer_units,
    remove_database,
    show_help,
)


if "--remove" in sys.argv:
    remove_timer_units()
    remove_database()
elif "--init" in sys.argv:
    add_timer_units()
elif "--config" in sys.argv:
    create_config()
    open_config()
elif "--help" in sys.argv:
    show_help()
else:
    if sys.argv[1:]:
        print(f"Error: Unknown argument: {sys.argv[1:]}")
        print("")
        show_help()
    else:
        main()



from .db import Db, ScheduleNotFound, ScheduleExhausted
from .config import Config, ConfigNotFound
from .helpers import get_notes_list, get_app_path, get_mail_content
from .mail import send_mail


def main() -> None:
    app_path = get_app_path()

    try:
        config = Config.load_from_file(app_path / "config.json")
    except ConfigNotFound:
        print(
            f"config.json not found at {app_path}, "
            "Please see README for more details!"
        )
        exit()

    db = Db((app_path / "zettel_merken.db"), config)

    db.create_schema()

    scheduled_notes = []

    for note in get_notes_list(config):
        if len(scheduled_notes) >= config.MAX_NOTES_PER_MAIL:
            break

        try:
            if db.is_note_scheduled(note):
                scheduled_notes.append(note)

        except ScheduleNotFound:
            db.create_note_schedule(note)

        except ScheduleExhausted:
            continue

    if scheduled_notes:
        mail_content = get_mail_content(scheduled_notes)
        send_mail(mail_content, config)
        db.update_note_schedule(scheduled_notes)
